/*
 * Controls the spawning, removal and pathing of the small buggers.
 * P.S. "Buggers" refer to the small worldlike characters of the game...
 */
#ifndef BUGGER_LOADER_H
#define BUGGER_LOADER_H

#include <float.h>

#include <vul_timer.h>

#include "slenderer.h"
#include "renderer/scene.h"
#include "renderer/quad.h"
#include "renderer/renderable.h"
#include "renderer/program.h"

#include "world_loader.h"

#define BUGGER_ANIMATION_HZ 6
#define BUGGER_WALK_SPEED_MS_PER_TILE 700 // @TODO: Tweak this as difficulty setting!
#define BUGGER_FALL_SPEED_MS_PER_TILE 250 // @TODO: This is linear which is really weird; do it differently?

#define LD_NUMBER_OF_SONGS 3

typedef enum ld_bugger_callback_type {
	LD_BUGGER_CALLBACK_SPAWN,
	LD_BUGGER_CALLBACK_REPATH,
	LD_BUGGER_CALLBACK_FALL,
	LD_BUGGER_CALLBACK_BOUNCE,
	LD_BUGGER_CALLBACK_WARP,
	LD_BUGGER_CALLBACK_EXIT,
	LD_BUGGER_CALLBACK_DIE,
	LD_BUGGER_CALLBACK_SOUND, // Play a sound. Sound id is given in bugger_id field
	LD_BUGGER_CALLBACK_COUNT
} ld_bugger_callback_type;

typedef struct ld_bugger {
	ui32_t id; // Doubles as index into ld_bugger_handler array.
	ui32_t quad_id;
	SL_BOOL direction_right;
} ld_bugger;

typedef struct ld_bugger_callback {
	ui64_t time;
	ui32_t bugger_id;
	ld_bugger_callback_type type;
} ld_bugger_callback;

typedef struct ld_bugger_handler {
	SL_BOOL simulation_running;
	ui64_t pause_delay, pause_start;
	ui32_t scene_id;
	ui32_t bugger_count;
	ui32_t bugger_dead_count;
	ui32_t bugger_exit_count;
	ld_bugger *buggers;
	vul_list_element_t *events; // Ordered list of callbacks; key is time to call the callback, content is the bugger_id
	
	// Shared by all the buggers
	sl_texture *bugger_tex;
	vul_vector_t *bugger_animation_states;
	ui32_t bugger_fall_clip_id,
		   bugger_bounce_clip_id,
		   bugger_enter_clip_id,
		   bugger_warp_clip_id,
		   bugger_exit_clip_id,
		   bugger_die_clip_id,
		   bugger_music_clip_id;
} ld_bugger_handler;

ld_bugger_handler *ld_bugger_handler_instance;

void ld_buggers_load( sl_scene *scene, ui32_t count, ui32_t spawn_pause, sl_vec *scale );

void ld_buggers_finalize( );

ui32_t ld_buggers_add( sl_scene *scene, sl_vec *scale );

void ld_bugger_update( );

int ld_bugger_callback_list_comparator( void *a, void *b );

void ld_bugger_spawn( ui64_t now, ui32_t id );

void ld_bugger_repath( ui64_t now, ui32_t id, SL_BOOL immediate_repath );

void ld_bugger_fall( ui64_t now, ui32_t id );

void ld_bugger_bounce( ui64_t now, ui32_t id, SL_BOOL immediate_rebounce );

void ld_bugger_warp( ui64_t now, ui32_t id );

void ld_bugger_die( ui32_t id );

void ld_bugger_exit( ui32_t id );

void ld_bugger_walk( const ld_bugger *bugger, const sl_quad *quad, sl_animator *a );

void ld_bugger_fall_anim( const ld_bugger *bugger, const sl_quad *quad, sl_animator *a, i32_t tile_count );

void ld_bugger_add_callback( ld_bugger_callback *cb );

void ld_bugger_sound( ui32_t id );

void ld_bugger_sound_stop( ui32_t id );

void ld_bugger_stop( );
void ld_bugger_start( );

#endif