/* 
 * Builds a dictionary from tile ids and textures.
 * Each type can have multiple textures.
 */
#ifndef TILE_DICTIONARY_H
#define TILE_DICTIONARY_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>

#include <vul_hash_map.h>
#include <vul_linked_list.h>
#define STBI_HEADER_FILE_ONLY
#include <stb_image.h>

#include "slenderer.h"
#include "renderer/texture.h"
#include "renderer/animator.h"

#ifdef _WIN32
	#define snprintf sprintf_s
#endif

#define TILE_DICTIONARY_BUCKET_COUNT 32
#define TILE_ANIMATION_FREQUENCY 4

#define TILE_INFINITELY_SWAPPABLE 0x7fffffff
#define TILE_PASSIVLEY_SWAPPABLE 0xffffffff

typedef struct ld_tile_dictionary_entry {
	ui32_t variation_count;
	ui32_t max_swaps;
	ui32_t *texture_ids;
	vul_vector_t **animation_states; // < sl_animation_sprite_state >
	ui32_t inverse_frequency; // Frames per second in animation
	// @TODO: Callbacks. Need down & up.
} ld_tile_dictionary_entry;

typedef struct ld_tile_dictionary {
	vul_list_element_t *keys; // ui32_t

	vul_hash_map_t *mapping; // <ui32_t, ld_tile_dictionary_entry >
} ld_tile_dictionary;

int ld_load_image_to_texture( sl_texture *tex_out, const char *path_to_image );

ld_tile_dictionary *ld_tile_dictionary_instance;

vul_list_element_t *ld_tile_dictionary_add_key( ui32_t key );

void ld_tile_dictionary_add_entry( vul_list_element_t *key_entry,
								   ui32_t var_count, ui32_t frame_count, ui32_t frame_freq,
								   ui32_t max_swaps, char *base_name );

void ld_tile_read_dictionary( );

void ld_tile_dictionary_finalize( );

ui32_t ld_ui32t_modulo_hash( const ui8_t *data, ui32_t len );

int ld_list_ui32t_comparator( void *a, void *b );

int ld_map_ui32t_comparator( void *a, void *b );

/*
 * Clean up the globals
 */
void ld_world_finalize( );

/*
 * Finds (random from selection) texture, animation states, animation frequency
 * and uvs for a given tile id.
 */
void ld_tile_dictionary_get( sl_texture **tex, vul_vector_t **anim_states, sl_box *uvs, f32_t *freq, ui32_t *max_swaps, ui32_t tile_id );

ui32_t ld_get_tile_swaps_by_texture_id( unsigned int tex_id );

#endif