/* 
 * Builds a dictionary from tile ids and textures.
 * Each type can have multiple textures.
 */
#ifndef WORLD_LOADER_H
#define WORLD_LOADER_H

#include <stdio.h>
#include <ctype.h>

#include "slenderer.h"
#include "renderer/scene.h"
#include "renderer/quad.h"
#include "renderer/renderable.h"
#include "renderer/program.h"

#include "tile_dictionary.h"

#define LD_WORLD_TILE_LAYER 0
#define LD_WORLD_BUGGER_LAYER 1
#define LD_WORLD_OVERLAY_LAYER 2
#define LD_WORLD_UI_LAYER 3

#define LD_WORLD_TILE_TYPE_WALL 0
#define LD_WORLD_TILE_TYPE_EARTH 1
#define LD_WORLD_TILE_TYPE_GROUND 2
#define LD_WORLD_TILE_TYPE_POROUS 3
#define LD_WORLD_TILE_TYPE_STATIC 4
#define LD_WORLD_TILE_TYPE_SPIKES 5
#define LD_WORLD_TILE_TYPE_TRAMPOLINE 6
#define LD_WORLD_TILE_TYPE_EARTH_SINGLE_SWAP 7
#define LD_WORLD_TILE_TYPE_ENTRY 8
#define LD_WORLD_TILE_TYPE_WARP_IN 9
#define LD_WORLD_TILE_TYPE_WARP_OUT 10
#define LD_WORLD_TILE_TYPE_EXIT 11

#define LD_NUMBER_OF_TRAMPOLINES 4

typedef struct ld_world_tile {
	ui32_t tile_type;
	ui32_t quad_id;
	ui32_t swaps_left;
} ld_world_tile;

typedef struct ld_world_state {
	ui32_t width, height;
	ld_world_tile *tiles;
	sl_vec entry, warp_out; // Position of entry and warp_out in normalized coords
	ui32_t trampolines_left;
	ui32_t *trampoline_ids;
} ld_world_state;

ld_world_state *ld_world_state_instance;

/*
 * Load a scene from a file.
 */
void ld_world_load_file( sl_scene *scene, const char *path );

/*
 * Reload a scene from file
 */
void ld_world_reload( sl_scene *scene, const char *path );

/*
 * Load a scene from an array of tile ids.
 */
void ld_world_load_array( sl_scene *scene, ui32_t **arrays, ui32_t width, ui32_t height, ui32_t trampolines );

/*
 * Loads a post-process program from a fragment program file.
 * Adds it to the scene.
 */
void ld_world_load_post_program( sl_scene *scene, const char *post_path );

/*
 * Create the world ui.
 */
void ld_world_ui_load( sl_scene *scene, ui32_t width, ui32_t height );

// Mouse callbacks
void ld_world_tile_over( unsigned int scene_id, unsigned int quad_id );
void ld_world_tile_out( unsigned int scene_id, unsigned int quad_id );
void ld_world_tile_swap( unsigned int scene_id, unsigned quad_id, int button );

void ld_world_normalized_coords_to_indices( ui32_t *ix, ui32_t *iy, sl_vec *normalized_pos );
void ld_world_indices_to_normalized_coords( sl_vec *normalized_pos, ui32_t ix, ui32_t iy );

ui32_t ld_world_swaps_remaining( sl_vec *normalized_pos );

void ld_world_swap( sl_scene *scene, sl_vec *normalized_pos );

ui32_t ld_world_get_quad_id_for_pos( sl_vec *normalized_pos );

void ld_world_change_tile( sl_scene *scene, ui32_t xi, ui32_t yi, ui32_t tile_id );

void ld_world_swap_tiles( sl_scene *scene, ui32_t ix, ui32_t iy, ui32_t ixb, ui32_t iyb );

void ld_world_insert_trampoline( sl_scene *scene, sl_vec *normalized_pos );

#endif
