#include "bugger_loader.h"
#include "ld30.h"
#include "bugger_loader.h"
#include "menu.h"

void ld_buggers_load( sl_scene *scene, ui32_t count, ui32_t spawn_pause, sl_vec *scale )
{
	ui32_t i;
	ui64_t now;
	sl_animation_sprite_state *as;
	sl_animator *a;
	sl_box uvs;
	sl_vec delta;
	vul_timer_t *clock;
	ld_bugger_callback cb;
	sl_aurator *aur;
	sl_aurator_clip *clip;
	char song_path[ 256 ];

	ld_bugger_handler_instance = ( ld_bugger_handler* )malloc( sizeof( ld_bugger_handler ) );
	ld_bugger_handler_instance->bugger_count = count;
	ld_bugger_handler_instance->buggers = ( ld_bugger* )malloc( sizeof( ld_bugger ) * count );
	ld_bugger_handler_instance->scene_id = scene->scene_id;
	ld_bugger_handler_instance->bugger_dead_count = 0;
	ld_bugger_handler_instance->bugger_exit_count = 0;
	ld_bugger_handler_instance->pause_delay = 0;
	ld_bugger_handler_instance->pause_start = 0;

	// Load the texture
	ld_bugger_handler_instance->bugger_tex = sl_renderer_allocate_texture( );
	ld_load_image_to_texture( ld_bugger_handler_instance->bugger_tex, "../res/buggers/bugger_walk.png" );
	
	// Set up the animation
	sl_vset( &delta, 0.5f, 0.5f );
	sl_bset_scalar( &uvs, 0.f, 0.f, 0.5f, 0.5f );
	ld_bugger_handler_instance->bugger_animation_states = vul_vector_create( sizeof( sl_animation_sprite_state ), 4 );
	for( i = 0; i < 4; ++i ) {
		as = ( sl_animation_sprite_state* )vul_vector_add_empty( ld_bugger_handler_instance->bugger_animation_states );
		as->texture_id = ld_bugger_handler_instance->bugger_tex->texture_id;
		sl_bset( &as->uvs, &uvs );
		sl_bset_scalar( &uvs, uvs.min_p.x + delta.x, uvs.min_p.y,
							  uvs.max_p.x + delta.x, uvs.max_p.y );
		if( uvs.min_p.x >= 0.99999f ) {
			sl_bset_scalar( &uvs, 0.f,	   uvs.min_p.y + delta.y,
								  delta.x, uvs.max_p.y + delta.y );
		}
	}

	// Create the buggers (hidden)
	a = sl_renderer_get_animator_for_scene( scene->scene_id );
	for( i = 0; i < count; ++i ) {
		ld_bugger_handler_instance->buggers[ i ].id = i;
		ld_bugger_handler_instance->buggers[ i ].direction_right = SL_TRUE;

		ld_bugger_handler_instance->buggers[ i ].quad_id = ld_buggers_add( scene, scale );

		sl_animator_add_sprite( a, ld_bugger_handler_instance->buggers[ i ].quad_id,
								   ld_bugger_handler_instance->bugger_animation_states,
								   ( ui64_t )( 1000 / BUGGER_ANIMATION_HZ ),
								   SL_ANIMATION_RUNNING_LOOPED );
	}

	// Create the sounds
	

	/* Add sound */	
	aur = sl_renderer_get_aurator_for_scene( scene->scene_id );
	ld_bugger_handler_instance->bugger_fall_clip_id = sl_aurator_allocate_clip( aur );
	clip = sl_aurator_get_clip( aur, ld_bugger_handler_instance->bugger_fall_clip_id);
	sl_aurator_load_ogg( aur, &clip, "../res/audio/fall.ogg" );
	clip->state = SL_AURATOR_STOPPED;
	
	ld_bugger_handler_instance->bugger_bounce_clip_id = sl_aurator_allocate_clip( aur );
	clip = sl_aurator_get_clip( aur, ld_bugger_handler_instance->bugger_bounce_clip_id);
	sl_aurator_load_ogg( aur, &clip, "../res/audio/bounce.ogg" );
	clip->state = SL_AURATOR_STOPPED;
	
	ld_bugger_handler_instance->bugger_enter_clip_id = sl_aurator_allocate_clip( aur );
	clip = sl_aurator_get_clip( aur, ld_bugger_handler_instance->bugger_enter_clip_id);
	sl_aurator_load_ogg( aur, &clip, "../res/audio/enter.ogg" );
	clip->state = SL_AURATOR_STOPPED;
	
	ld_bugger_handler_instance->bugger_exit_clip_id = sl_aurator_allocate_clip( aur );
	clip = sl_aurator_get_clip( aur, ld_bugger_handler_instance->bugger_exit_clip_id);
	sl_aurator_load_ogg( aur, &clip, "../res/audio/exit.ogg" );
	clip->state = SL_AURATOR_STOPPED;
	
	ld_bugger_handler_instance->bugger_warp_clip_id = sl_aurator_allocate_clip( aur );
	clip = sl_aurator_get_clip( aur, ld_bugger_handler_instance->bugger_warp_clip_id);
	sl_aurator_load_ogg( aur, &clip, "../res/audio/warp.ogg" );
	clip->state = SL_AURATOR_STOPPED;
	
	ld_bugger_handler_instance->bugger_die_clip_id = sl_aurator_allocate_clip( aur );
	clip = sl_aurator_get_clip( aur, ld_bugger_handler_instance->bugger_die_clip_id);
	sl_aurator_load_ogg( aur, &clip, "../res/audio/die.ogg" );
	clip->state = SL_AURATOR_STOPPED;
	
	i = ( ui32_t )( ( f32_t )LD_NUMBER_OF_SONGS * ( ( f32_t )rand( ) / ( f32_t )RAND_MAX ) );
	snprintf( song_path, 255, "../res/audio/music%d.ogg", i );
	ld_bugger_handler_instance->bugger_music_clip_id = sl_aurator_allocate_clip( aur );
	clip = sl_aurator_get_clip( aur, ld_bugger_handler_instance->bugger_music_clip_id );
	sl_aurator_load_ogg( aur, &clip, song_path );
	clip->state = SL_AURATOR_RUNNING_LOOPED;
	
	// Finally, create the callbacks that spawn them
	clock = a->clock;
	now = vul_timer_get_millis( clock );
	ld_bugger_handler_instance->events = NULL;
	for( i = 0; i < count; ++i ) {
		cb.time = now + ( ui64_t )( i + 1 ) * ( ui64_t )spawn_pause;
		cb.bugger_id = ld_bugger_handler_instance->buggers[ i ].id;
		cb.type = LD_BUGGER_CALLBACK_SPAWN;
		ld_bugger_add_callback( &cb );
	}

	// And set us as running
	ld_bugger_handler_instance->simulation_running = SL_TRUE;
}

void ld_buggers_finalize( )
{
	if( ld_bugger_handler_instance ) {
		vul_vector_destroy( ld_bugger_handler_instance->bugger_animation_states );
		if( ld_bugger_handler_instance->events ) {
			vul_list_destroy( ld_bugger_handler_instance->events );
		}
		free( ld_bugger_handler_instance->buggers );
		free( ld_bugger_handler_instance );
	}
}

ui32_t ld_buggers_add( sl_scene *scene, sl_vec *scale )
{
	sl_box uvs;
	GLfloat color[ 4 ];
	sl_bvec flip;

	sl_bset_scalar( &uvs, 0.f, 0.f, 0.5f, 0.5f ); // Buggers is a 2x2 texture
	color[ 0 ] = 1.f; color[ 1 ] = 1.f; color[ 2 ] = 1.f; color[ 3 ] = 1.f;
	flip.x = SL_FALSE; flip.y = SL_FALSE;
	return sl_scene_add_sprite( scene, LD_WORLD_BUGGER_LAYER,
								&ld_world_state_instance->entry, scale, 0.f,
								ld_bugger_handler_instance->bugger_tex->texture_id,
								ld_program_instance_id, 
								ld_renderable_instance_id,
								&uvs, &flip, color, SL_TRUE ); // Hidden by default
}

void ld_bugger_add_callback( ld_bugger_callback *cb )
{
	vul_list_element_t *e;

	e = vul_list_insert( ld_bugger_handler_instance->events, cb, sizeof( ld_bugger_callback ), ld_bugger_callback_list_comparator );
	if( e->prev == NULL ) {
		ld_bugger_handler_instance->events = e; // New head of the list.
	}
}

void ld_bugger_update( )
{
	ui64_t now;
	sl_animator *animator;
	sl_scene *ls, *ms;
	vul_list_element_t *e;
	ld_bugger_callback *cb;
	char level_path[ 256 ];

	// If not running, return
	if( !ld_bugger_handler_instance
		|| !ld_bugger_handler_instance->simulation_running ) {
		return;
	}
	animator = sl_renderer_get_animator_for_scene( ld_bugger_handler_instance->scene_id );

	// Timing
	now = vul_timer_get_millis( animator->clock );
	// Adjust now for paused time
	now -= ld_bugger_handler_instance->pause_delay;
	
	while( ( e = ld_bugger_handler_instance->events ) ) {
		assert( e->data_size == sizeof( ld_bugger_callback ) );
		
		cb = ( ld_bugger_callback* )e->data;
		
		if( cb->time > now ) {
			// We have a sorted list, so this means we can break
			break;
		}

		// There is a callback that should be called
		switch( cb->type ) {
		case LD_BUGGER_CALLBACK_SPAWN:
			ld_bugger_spawn( now, cb->bugger_id );
			break;
		case LD_BUGGER_CALLBACK_REPATH:
			ld_bugger_repath( now, cb->bugger_id, SL_FALSE );
			break;
		case LD_BUGGER_CALLBACK_FALL:
			ld_bugger_fall( now, cb->bugger_id );
			break;
		case LD_BUGGER_CALLBACK_BOUNCE:
			ld_bugger_bounce( now, cb->bugger_id, SL_FALSE );
			break;
		case LD_BUGGER_CALLBACK_WARP:
			ld_bugger_warp( now, cb->bugger_id );
			break;
		case LD_BUGGER_CALLBACK_EXIT:
			ld_bugger_exit( cb->bugger_id );
			break;
		case LD_BUGGER_CALLBACK_DIE:
			ld_bugger_die( cb->bugger_id );
			break;
		case LD_BUGGER_CALLBACK_SOUND:
			ld_bugger_sound( cb->bugger_id );
			break;
		case LD_BUGGER_CALLBACK_COUNT:
			assert( SL_FALSE ); // Not a valid callback type!
			break;
		}
		// And remove it (this is from the front, so no need to update much
		ld_bugger_handler_instance->events = e->next;
		if( ld_bugger_handler_instance->events ) {
			ld_bugger_handler_instance->events->prev = NULL; // Only update this one...
		}
		free( e->data );
		free( e );
	}

	
	if( ld_bugger_handler_instance->bugger_exit_count
		== ld_bugger_handler_instance->bugger_count ) {
		// Level over, you won
		// ld_bugger_handler_instance->simulation_running = SL_FALSE;
		ld_bugger_sound_stop( ld_bugger_handler_instance->bugger_music_clip_id );
		printf( "Level %d completed.\n", ld_game_current_level );
		if ( ld_game_current_level  == 0xffffffff ) {
			printf( "You beat the custom levvel!\n" );
				
			ms = sl_renderer_get_scene_by_id( ld_menu_scene_id );
			assert( ms );
	
			ld_last_world_scene_id = ld_active_scene_id;
			ld_active_scene_id = ld_menu_scene_id;	
			// Change play button behaviour
			sl_controller_add_mouse_up_callback( ld_menu_scene_id, ld_menu_play_button, ld_resume );
			ld_bugger_stop( );
			return;
		} else if( ++ld_game_current_level < LD_LEVELS_IMPLEMENTED ) {
				
			snprintf( level_path, 255, "../res/levels/lvl%d.scn", ld_game_current_level );
			printf( "Loading %s\n", level_path );
				
			ls = sl_renderer_get_scene_by_id( ld_active_scene_id );
			assert( ls );

			ld_world_reload( ls, level_path );

			ld_bugger_start( );
			return;
		} else {
			printf( "We're out of levels, laod custom ones instead!\n" );
				
			ld_game_current_level = 0; // Reset to first level

			ms = sl_renderer_get_scene_by_id( ld_menu_scene_id );
			assert( ms );
		
			ld_last_world_scene_id = ld_active_scene_id;
			ld_active_scene_id = ld_menu_scene_id;	
			// Change play button behaviour
			sl_controller_add_mouse_up_callback( ld_menu_scene_id, ld_menu_play_button, ld_resume );
			ld_bugger_stop( );
			return;
		}
	}
	
	if( ld_bugger_handler_instance->bugger_dead_count +
		ld_bugger_handler_instance->bugger_exit_count
		== ld_bugger_handler_instance->bugger_count ) {
			// Level is over but you lost, reload
			printf( "You failed the level, so try again!\n" );
			
			snprintf( level_path, 255, "../res/levels/lvl%d.scn", ld_game_current_level );
			printf( "Reloading %s\n", level_path );
				
			ls = sl_renderer_get_scene_by_id( ld_active_scene_id );
			assert( ls );

			ld_world_reload( ls, level_path );

			ld_bugger_start( );
			return;
	}
}

int ld_bugger_callback_list_comparator( void *a, void *b )
{
	ui64_t time_a, time_b;

	time_a = *( ( ui64_t* )a );
	time_b = *( ( ui64_t* )b );

	return ( int )( ( i64_t )time_a - ( i64_t )time_b );
}

void ld_bugger_spawn( ui64_t now, ui32_t id )
{
	ld_bugger *b;
	sl_scene *scene;
	sl_quad *q;
	
	// Get the bugger
	scene = sl_renderer_get_scene_by_id( ld_bugger_handler_instance->scene_id );
	b = &ld_bugger_handler_instance->buggers[ id ];

	// Make it visible
	q = sl_scene_get_volitile_quad( scene, b->quad_id, LD_WORLD_BUGGER_LAYER );
	q->hidden = SL_FALSE;

	// Immediately repath
	ld_bugger_repath( now, id, SL_FALSE );

	// Play spawn sound
	ld_bugger_sound( ld_bugger_handler_instance->bugger_enter_clip_id );
}

void ld_bugger_repath( ui64_t now, ui32_t id, SL_BOOL immediate_repath )
{
	ui32_t xi, yi, xnexti;
	ld_bugger *b;
	sl_scene *scene;
	sl_animator *animator;
	sl_vec pos;
	const sl_quad *q;
	sl_quad *qv;
	ld_world_tile *tile;
	ld_bugger_callback cb;
	
	// Get the bugger
	scene = sl_renderer_get_scene_by_id( ld_bugger_handler_instance->scene_id );
	animator = sl_renderer_get_animator_for_scene( ld_bugger_handler_instance->scene_id );	
	b = &ld_bugger_handler_instance->buggers[ id ];
	q = sl_scene_get_const_quad( scene, b->quad_id, LD_WORLD_BUGGER_LAYER );

	// Calculate indices for this its and the next one
	sl_vset( &pos, q->world_matrix.data[ 12 ], q->world_matrix.data[ 13 ] );
	ld_world_normalized_coords_to_indices( &xi, &yi, &pos );
	xnexti = b->direction_right ? xi + 1 : xi - 1;

	// If the next tile is outside the world, turn around
	if( xnexti >= ld_world_state_instance->width ) {
		xnexti = xi - 1;
		b->direction_right = SL_FALSE;
	} else if( xi == 0 && !b->direction_right ) {
		xnexti = xi + 1;
		b->direction_right = SL_TRUE;
	}
	
	tile = &ld_world_state_instance->tiles[ yi * ld_world_state_instance->width + xnexti ];
	
	switch( tile->tile_type ) {
	case LD_WORLD_TILE_TYPE_GROUND:
	case LD_WORLD_TILE_TYPE_ENTRY:
	case LD_WORLD_TILE_TYPE_WARP_OUT:
		// If the next tile is pathable, just walk. Pathable tiles are: ground, porous, entry, warp_out
		ld_bugger_walk( b, q, animator );
		cb.bugger_id = b->id;
		cb.type = LD_BUGGER_CALLBACK_REPATH;
		cb.time = now + BUGGER_WALK_SPEED_MS_PER_TILE;
		ld_bugger_add_callback( &cb );
		break;
	case LD_WORLD_TILE_TYPE_POROUS:
		assert( SL_FALSE ); // We haven't actually implemented porous tiles yet!
		// @TODO: Walk there, remember that porous tiles should break after a few walks.
		// If we exausted it add POROUS_EXHAUSTED callback in 1 WALK_HZ
		// @TODO: Wait with porous until after the rest works
		break;
	case LD_WORLD_TILE_TYPE_WALL:
		// Walk into the air, then _FALL
		ld_bugger_walk( b, q, animator );
		cb.bugger_id = b->id;
		cb.type = LD_BUGGER_CALLBACK_FALL;
		cb.time = now + BUGGER_WALK_SPEED_MS_PER_TILE;
		ld_bugger_add_callback( &cb );
		break;
	case LD_WORLD_TILE_TYPE_SPIKES:
		// Walk once, callback of type _DIE
		ld_bugger_walk( b, q, animator );
		cb.bugger_id = b->id;
		cb.type = LD_BUGGER_CALLBACK_DIE;
		cb.time = now + BUGGER_WALK_SPEED_MS_PER_TILE;
		ld_bugger_add_callback( &cb );
		break;
	case LD_WORLD_TILE_TYPE_WARP_IN:
		// Walk once, callback is _WARP
		ld_bugger_walk( b, q, animator );
		cb.bugger_id = b->id;
		cb.type = LD_BUGGER_CALLBACK_WARP;
		cb.time = now + BUGGER_WALK_SPEED_MS_PER_TILE;
		ld_bugger_add_callback( &cb );
		break;
	case LD_WORLD_TILE_TYPE_EXIT:
		// Walk once, callback is _EXIT
		ld_bugger_walk( b, q, animator );
		cb.bugger_id = b->id;
		cb.type = LD_BUGGER_CALLBACK_EXIT;
		cb.time = now + BUGGER_WALK_SPEED_MS_PER_TILE;
		ld_bugger_add_callback( &cb );
		break;
	case LD_WORLD_TILE_TYPE_TRAMPOLINE:
		// Walk into the trampoline, then _BOUNCE
		ld_bugger_walk( b, q, animator );
		cb.bugger_id = b->id;
		cb.type = LD_BUGGER_CALLBACK_BOUNCE;
		cb.time = now + BUGGER_WALK_SPEED_MS_PER_TILE;
		ld_bugger_add_callback( &cb );
		// Flip the trampoline for style
		if( !b->direction_right ) {
			qv = sl_scene_get_volitile_quad( scene, tile->quad_id, LD_WORLD_TILE_LAYER );
			qv->flip_uvs.x = SL_TRUE;
		} else {
			qv = sl_scene_get_volitile_quad( scene, tile->quad_id, LD_WORLD_TILE_LAYER );
			qv->flip_uvs.x = SL_FALSE;
		}
		break;
	case LD_WORLD_TILE_TYPE_EARTH:
	case LD_WORLD_TILE_TYPE_EARTH_SINGLE_SWAP:
	case LD_WORLD_TILE_TYPE_STATIC:
		// Turn around
		b->direction_right = b->direction_right ? SL_FALSE : SL_TRUE;
		qv = sl_scene_get_volitile_quad( scene, b->quad_id, LD_WORLD_BUGGER_LAYER );
		qv->flip_uvs.x = qv->flip_uvs.x ? SL_FALSE : SL_TRUE;
		// Try prepathing immediately if not already attempted
		if( !immediate_repath ) {
			ld_bugger_repath( now, id, SL_TRUE );
		} else {
			// Repath again in 1 tile of waiting; this is to avaid infinite recursion if walls are on both sides.
			cb.bugger_id = b->id;
			cb.type = LD_BUGGER_CALLBACK_REPATH;
			cb.time = now + BUGGER_WALK_SPEED_MS_PER_TILE;
			ld_bugger_add_callback( &cb );
			break;
		}
		break;
	}
}

void ld_bugger_fall( ui64_t now, ui32_t id )
{
	// We need to figure out what we're falling onto, and then it's pretty easy...
	ui32_t xi, yi, ynexti;
	ld_bugger *b;
	sl_scene *scene;
	sl_animator *animator;
	sl_vec pos;
	const sl_quad *q;
	ld_world_tile *tile;
	ld_bugger_callback cb;

	// Get the bugger
	scene = sl_renderer_get_scene_by_id( ld_bugger_handler_instance->scene_id );
	animator = sl_renderer_get_animator_for_scene( ld_bugger_handler_instance->scene_id );	
	b = &ld_bugger_handler_instance->buggers[ id ];
	q = sl_scene_get_const_quad( scene, b->quad_id, LD_WORLD_BUGGER_LAYER );
	
	// Calculate indices for this tile
	sl_vset( &pos, q->world_matrix.data[ 12 ], q->world_matrix.data[ 13 ] );
	ld_world_normalized_coords_to_indices( &xi, &yi, &pos );
	// Find the next one below that isn't a wall
	ynexti = yi;
	while( ++ynexti < ld_world_state_instance->height ) {
		tile = &ld_world_state_instance->tiles[ ynexti * ld_world_state_instance->width + xi ];
		switch( tile->tile_type ) {
		case LD_WORLD_TILE_TYPE_WALL:
			continue; // This is the only case in which we kee goign down
		case LD_WORLD_TILE_TYPE_EARTH:
		case LD_WORLD_TILE_TYPE_EARTH_SINGLE_SWAP:
		case LD_WORLD_TILE_TYPE_STATIC:
			// We have created a hole here, so just replace the tile above it with grass and we're fine!
			ld_world_change_tile( scene, xi, --ynexti, LD_WORLD_TILE_TYPE_GROUND );
			// Then behave as for grass!
		// FALL THOUGH
		case LD_WORLD_TILE_TYPE_GROUND:
		case LD_WORLD_TILE_TYPE_ENTRY:
		case LD_WORLD_TILE_TYPE_WARP_OUT:
		case LD_WORLD_TILE_TYPE_POROUS:
			// All of these are valid walking-tiles, so just fall to them and _REPATH
			ld_bugger_fall_anim( b, q, animator, ( i32_t )( ynexti - yi ) );
			cb.bugger_id = b->id;
			cb.type = LD_BUGGER_CALLBACK_REPATH;
			cb.time = now + BUGGER_FALL_SPEED_MS_PER_TILE * ( ynexti - yi );
			ld_bugger_add_callback( &cb );
			cb.bugger_id = ld_bugger_handler_instance->bugger_fall_clip_id; // Add falling sound
			cb.type = LD_BUGGER_CALLBACK_SOUND;
			cb.time = now + BUGGER_FALL_SPEED_MS_PER_TILE * ( ynexti - yi );
			ld_bugger_add_callback( &cb );
			break;
		case LD_WORLD_TILE_TYPE_TRAMPOLINE:
			// Fall, then bounce back up, so just fall to them and _BOUNCE
			ld_bugger_fall_anim( b, q, animator, ( i32_t )( ynexti - yi ) );
			cb.bugger_id = b->id;
			cb.type = LD_BUGGER_CALLBACK_BOUNCE;
			cb.time = now + BUGGER_FALL_SPEED_MS_PER_TILE * ( ynexti - yi );
			ld_bugger_add_callback( &cb );
			cb.bugger_id = ld_bugger_handler_instance->bugger_bounce_clip_id; // Add bouncing sound
			cb.type = LD_BUGGER_CALLBACK_SOUND;
			cb.time = now + BUGGER_FALL_SPEED_MS_PER_TILE * ( ynexti - yi );
			ld_bugger_add_callback( &cb );
			break;
		case LD_WORLD_TILE_TYPE_WARP_IN:
			// Fall to it, then _WARP
			ld_bugger_fall_anim( b, q, animator, ( i32_t )( ynexti - yi ) );
			cb.bugger_id = b->id;
			cb.type = LD_BUGGER_CALLBACK_WARP;
			cb.time = now + BUGGER_FALL_SPEED_MS_PER_TILE * ( ynexti - yi );
			ld_bugger_add_callback( &cb );
			break;
		case LD_WORLD_TILE_TYPE_EXIT:
			// Fall to it, then _EXIT
			ld_bugger_fall_anim( b, q, animator, ( i32_t )( ynexti - yi ) );
			cb.bugger_id = b->id;
			cb.type = LD_BUGGER_CALLBACK_EXIT;
			cb.time = now + BUGGER_FALL_SPEED_MS_PER_TILE * ( ynexti - yi );
			ld_bugger_add_callback( &cb );
			break;
		case LD_WORLD_TILE_TYPE_SPIKES:
			// Fall to it, then _DIE
			ld_bugger_fall_anim( b, q, animator, ( i32_t )( ynexti - yi ) );
			cb.bugger_id = b->id;
			cb.type = LD_BUGGER_CALLBACK_DIE;
			cb.time = now + BUGGER_FALL_SPEED_MS_PER_TILE * ( ynexti - yi );
			ld_bugger_add_callback( &cb );
			break;
		default:
			assert( SL_FALSE ); // Invalid tile type
		}
		break;
	}
}


void ld_bugger_bounce( ui64_t now, ui32_t id, SL_BOOL immediate_rebounce )
{
	// We need to figure out what we're bouncing onto, and then it's pretty easy...
	ui32_t xi, yi;
	i32_t ynexti, xnexti;
	ld_bugger *b;
	sl_scene *scene;
	sl_animator *animator;
	sl_vec pos;
	sl_quad *q;
	ld_world_tile *tile_above, *tile_next;
	ld_bugger_callback cb;

	// Get the bugger
	scene = sl_renderer_get_scene_by_id( ld_bugger_handler_instance->scene_id );
	animator = sl_renderer_get_animator_for_scene( ld_bugger_handler_instance->scene_id );	
	b = &ld_bugger_handler_instance->buggers[ id ];
	q = sl_scene_get_volitile_quad( scene, b->quad_id, LD_WORLD_BUGGER_LAYER );
	
	// Calculate indices for this tile
	sl_vset( &pos, q->world_matrix.data[ 12 ], q->world_matrix.data[ 13 ] );
	ld_world_normalized_coords_to_indices( &xi, &yi, &pos );
	// Find the next one below that isn't a wall
	ynexti = ( i32_t )yi;
	xnexti = ( i32_t )xi + ( b->direction_right ? 1 : -1 );
	if( ( xi == 0 && !b->direction_right ) || ( xnexti == ld_world_state_instance->width && b->direction_right ) ) {
		// We can't jump, try other direction then turn and repath
		b->direction_right = b->direction_right ? SL_FALSE : SL_TRUE;
		if( immediate_rebounce ) {
			b->direction_right = b->direction_right ? SL_FALSE : SL_TRUE; // The second turn means we keep trying in the same direction we originally went, and only turn if there's a wall
			ld_bugger_repath( now, id, SL_FALSE ); 
		} else {
			ld_bugger_bounce( now, id, SL_TRUE );
		}
	}
	while( --ynexti >= 0 ) {
		tile_above = &ld_world_state_instance->tiles[ ynexti * ld_world_state_instance->width + xi ];
		tile_next = &ld_world_state_instance->tiles[ ynexti * ld_world_state_instance->width + xnexti ];
		
		if( tile_above->tile_type != LD_WORLD_TILE_TYPE_WALL ) {
			// We have hit a roof of some kind before finding an exit.
			// Try the other direction, otherwise turn and repath
			b->direction_right = b->direction_right ? SL_FALSE : SL_TRUE;
			if( immediate_rebounce ) {
				ld_bugger_repath( now, id, SL_FALSE ); 
			} else {
				ld_bugger_bounce( now, id, SL_TRUE );
			}
			q->flip_uvs.x = !b->direction_right;
			break;
		}
		if( tile_next->tile_type == LD_WORLD_TILE_TYPE_GROUND
			|| tile_next->tile_type == LD_WORLD_TILE_TYPE_WARP_IN
			|| tile_next->tile_type == LD_WORLD_TILE_TYPE_WARP_OUT
			|| tile_next->tile_type == LD_WORLD_TILE_TYPE_EXIT
			|| tile_next->tile_type == LD_WORLD_TILE_TYPE_POROUS
			|| tile_next->tile_type == LD_WORLD_TILE_TYPE_TRAMPOLINE ) {
				// We have found our exit. Bounce animation, then repath
			ld_bugger_fall_anim( b, q, animator, ynexti - yi );
			cb.bugger_id = b->id;
			cb.type = LD_BUGGER_CALLBACK_REPATH;
			cb.time = now + BUGGER_FALL_SPEED_MS_PER_TILE * abs( ynexti - yi );
			ld_bugger_add_callback( &cb );
			cb.bugger_id = ld_bugger_handler_instance->bugger_bounce_clip_id; // Add bouncing sound
			cb.type = LD_BUGGER_CALLBACK_SOUND;
			cb.time = now + BUGGER_FALL_SPEED_MS_PER_TILE * abs( ynexti - yi );
			ld_bugger_add_callback( &cb );
			break;
		}
	}
}

void ld_bugger_warp( ui64_t now, ui32_t id )
{
	sl_scene *scene;
	ld_bugger *b;
	sl_quad *q;

	scene = sl_renderer_get_scene_by_id( ld_bugger_handler_instance->scene_id );

	b = &ld_bugger_handler_instance->buggers[ id ];
	q = sl_scene_get_volitile_quad( scene, b->quad_id, LD_WORLD_BUGGER_LAYER );

	// Move the bugger to the warp zone
	q->world_matrix.data[ 12 ] = ld_world_state_instance->warp_out.x;
	q->world_matrix.data[ 13 ] = ld_world_state_instance->warp_out.y;

	// Now repath
	ld_bugger_repath( now, id, SL_FALSE );

	// Play the warp sound
	ld_bugger_sound( ld_bugger_handler_instance->bugger_warp_clip_id );
}

void ld_bugger_die( ui32_t id )
{
	sl_scene *scene;
	ld_bugger *b;
	sl_quad *q;

	scene = sl_renderer_get_scene_by_id( ld_bugger_handler_instance->scene_id );

	b = &ld_bugger_handler_instance->buggers[ id ];
	q = sl_scene_get_volitile_quad( scene, b->quad_id, LD_WORLD_BUGGER_LAYER );

	q->hidden = SL_TRUE;

	// Increase the death toll
	++ld_bugger_handler_instance->bugger_dead_count;

	// Play the death sound
	ld_bugger_sound( ld_bugger_handler_instance->bugger_die_clip_id );
}

void ld_bugger_exit( ui32_t id )
{
	sl_scene *scene;
	ld_bugger *b;
	sl_quad *q;

	scene = sl_renderer_get_scene_by_id( ld_bugger_handler_instance->scene_id );

	b = &ld_bugger_handler_instance->buggers[ id ];
	q = sl_scene_get_volitile_quad( scene, b->quad_id, LD_WORLD_BUGGER_LAYER );

	q->hidden = SL_TRUE;

	// Increase the exit toll
	++ld_bugger_handler_instance->bugger_exit_count;

	// Play the exit sound
	ld_bugger_sound( ld_bugger_handler_instance->bugger_exit_clip_id );
}

void ld_bugger_walk( const ld_bugger *bugger, const sl_quad *quad, sl_animator *a )
{
	sl_mat4 mat;
	sl_vec delta;

	// Calculate the end world matrix
	sl_vset( &delta, 2.f / ( f32_t )ld_world_state_instance->width,
					 2.f / ( f32_t )ld_world_state_instance->height );
	sl_mcopy4( &mat, &quad->world_matrix );
	if( bugger->direction_right ) {
		mat.data[ 12 ] += delta.x;
	} else {
		mat.data[ 12 ] -= delta.x;
	}

	// Add the animation
	sl_animator_add_transform( a, quad->quad_id, &mat, BUGGER_WALK_SPEED_MS_PER_TILE, SL_ANIMATION_RUNNING );
}

void ld_bugger_fall_anim( const ld_bugger *bugger, const sl_quad *quad, sl_animator *a, i32_t tile_count )
{
	sl_mat4 mat;
	sl_vec delta;

	// Calculate the end world matrix
	sl_vset( &delta, 2.f / ( f32_t )ld_world_state_instance->width,
					 2.f / ( f32_t )ld_world_state_instance->height );
	sl_mcopy4( &mat, &quad->world_matrix );
	mat.data[ 13 ] -= delta.y * tile_count;
	
	// Add the animation
	sl_animator_add_transform( a, quad->quad_id, &mat, BUGGER_FALL_SPEED_MS_PER_TILE * abs( tile_count ), SL_ANIMATION_RUNNING );
}

void ld_bugger_stop( )
{
	sl_animator *a;

	if( !ld_bugger_handler_instance ) {
		return; // @TODO: Maybe make this an assert
	}
	
	a = sl_renderer_get_animator_for_scene( ld_active_scene_id );
	ld_bugger_handler_instance->pause_start = vul_timer_get_millis( a->clock );
	ld_bugger_handler_instance->simulation_running = SL_FALSE;
}
void ld_bugger_start( )
{
	sl_animator *a;

	if( !ld_bugger_handler_instance ) {
		return; // @TODO: Maybe make this an assert
	}

	a = sl_renderer_get_animator_for_scene( ld_active_scene_id );
	if( ld_bugger_handler_instance->pause_start != 0 ) {
		ld_bugger_handler_instance->pause_delay +=
			vul_timer_get_millis( a->clock ) - ld_bugger_handler_instance->pause_start;
	}
	ld_bugger_handler_instance->simulation_running = SL_TRUE;
}

void ld_bugger_sound( ui32_t id )
{
	sl_aurator *aur;
	sl_aurator_clip *clip;
	
	aur = sl_renderer_get_aurator_for_scene( ld_bugger_handler_instance->scene_id );
	clip = sl_aurator_get_clip( aur, id );

	if( id == ld_bugger_handler_instance->bugger_fall_clip_id 
		|| id == ld_bugger_handler_instance->bugger_bounce_clip_id
		|| id == ld_bugger_handler_instance->bugger_warp_clip_id
		|| id == ld_bugger_handler_instance->bugger_enter_clip_id
		|| id == ld_bugger_handler_instance->bugger_exit_clip_id
		|| id == ld_bugger_handler_instance->bugger_die_clip_id ) {
		
		clip->state = SL_AURATOR_RUNNING_KEEP;
		clip->sample_current = 0;
	} else {
		assert( SL_FALSE ); // unknown clip id
	}
}

void ld_bugger_sound_stop( ui32_t id )
{
	sl_aurator *aur;
	sl_aurator_clip *clip;
	
	aur = sl_renderer_get_aurator_for_scene( ld_bugger_handler_instance->scene_id );
	clip = sl_aurator_get_clip( aur, id );

	if( clip != NULL ) {
		clip->state = SL_AURATOR_STOPPED;
	}
}
