#include "menu.h"

#include "bugger_loader.h"

#define STBI_HEADER_FILE_ONLY
#include "stb_image.h"

#define ROOMS_ID_MENU 0

void ld_open_menu( GLFWwindow *win_handle, int key, int scancode, int modifiers )
{
	sl_scene *s;

	s = sl_renderer_get_scene_by_id( ld_menu_scene_id );

	assert( s );
	ld_active_scene_id = s->scene_id;
}

void ld_toggle_menu( GLFWwindow *win_handle, int key, int scancode, int modifiers )
{
	sl_scene *s;

	s = sl_renderer_get_scene_by_id( ld_menu_scene_id );

	assert( s );
	
	ld_last_world_scene_id = ld_active_scene_id;
	ld_active_scene_id = ld_menu_scene_id;	
	// Change play button behaviour
	sl_controller_add_mouse_up_callback( s->scene_id, ld_menu_play_button, ld_resume );
	ld_bugger_stop( );
}

void ld_resume( unsigned int scene_id, unsigned int quad_id, int button )
{
	sl_scene *s;

	s = sl_renderer_get_scene_by_id( ld_last_world_scene_id );
	
	if( scene_id != ld_menu_scene_id ) {
		return; // Ignore it, not for this scene
	}

	assert( s );
	ld_active_scene_id = ld_last_world_scene_id;	
	ld_bugger_start( );
}

void ld_quit_mouse( unsigned int scene_id, unsigned int quad_id, int button )
{
	sl_window *w;

	if( scene_id != ld_active_scene_id ) {
		return; // Ignore it, not for this scene
	}

	w = sl_renderer_get_window_by_id( ld_window_instance_id );
	glfwSetWindowShouldClose( w->handle, GL_TRUE );
}

void ld_play( unsigned int scene_id, unsigned int quad_id, int button )
{
	sl_scene *scene;
	char level_path[ 256 ];

	if( scene_id != ld_active_scene_id ) {
		return; // Ignore it, not for this scene
	}
	// Create the game world; load all the rooms, make first room active.
	// Keep this room around for when menu is selected!
	scene = sl_renderer_add_scene( ld_window_instance_id, 0 ); // Post program id is irrelevant, is replaced in
								   // ld_world_load_file anyway.
	snprintf( level_path, 255, "../res/levels/lvl%d.scn", ld_game_current_level );
	printf( "Loading %s\n", level_path );
	ld_world_load_file( scene, level_path );
	ld_active_scene_id = scene->scene_id;
	
	sl_controller_add_key_press_callback( GLFW_KEY_ESCAPE, ld_toggle_menu );
	sl_controller_add_key_press_callback( GLFW_KEY_R, ld_reload );

	ld_bugger_start( );
}

void ld_play_custom( unsigned int scene_id, unsigned int quad_id, int button )
{
	sl_scene *scene;

	if( scene_id != ld_active_scene_id ) {
		return; // Ignore it, not for this scene
	}
	// Create the game world; load all the rooms, make first room active.
	// Keep this room around for when menu is selected!
	scene = sl_renderer_add_scene( ld_window_instance_id, 0 ); // Post program id is irrelevant, is replaced in
								   // ld_world_load_file anyway.
	ld_game_current_level = 0xffffffff;
	printf( "Loading %s\n", ld_game_custom_level_path );
	ld_world_load_file( scene, ld_game_custom_level_path );
	ld_active_scene_id = scene->scene_id;
	
	sl_controller_add_key_press_callback( GLFW_KEY_ESCAPE, ld_toggle_menu );
	sl_controller_add_key_press_callback( GLFW_KEY_R, ld_reload );

	ld_bugger_start( );
}

void ld_reload( GLFWwindow *win_handle, int key, int scancode, int modifiers )
{
	char level_path[ 256 ];
	sl_scene *s;

	s = sl_renderer_get_scene_by_id( ld_active_scene_id );
	// Create the game world; load all the rooms, make first room active.
	// Keep this room around for when menu is selected!
	snprintf( level_path, 255, "../res/levels/lvl%d.scn", ld_game_current_level );
	printf( "Reloding %s\n", level_path );
	ld_world_reload( s, level_path );

	sl_controller_add_key_press_callback( GLFW_KEY_ESCAPE, ld_toggle_menu );
	ld_bugger_start( );
}

void ld_toggle_credits( unsigned int scene_id, unsigned int quad_id, int button )
{
	sl_quad *q;
	sl_scene *s;
	
	s = sl_renderer_get_scene_by_id( ld_menu_scene_id );
	assert( s );

	q = sl_scene_get_volitile_quad( s, ld_menu_credits_id, 0xffffffff );
	q->hidden = q->hidden ? SL_FALSE : SL_TRUE;
}
void ld_toggle_help( unsigned int scene_id, unsigned int quad_id, int button )
{
	sl_quad *q;
	sl_scene *s;
	
	s = sl_renderer_get_scene_by_id( ld_menu_scene_id );
	assert( s );

	q = sl_scene_get_volitile_quad( s, ld_menu_help_id, 0xffffffff );
	q->hidden = q->hidden ? SL_FALSE : SL_TRUE;
}

void ld_menu_destroy( )
{
	vul_timer_destroy( ld_menu_post_timer );
}

void ld_menu_create( sl_window *win )
{
	sl_scene *s;
	sl_texture *t;
	sl_program *post_program;
	sl_aurator *ar;
	sl_aurator_clip *clip;
	sl_vec pos, scale;
	sl_box uvs, aabb;
	sl_bvec flip;
	GLfloat color[ 4 ];
	unsigned int id;

	// Set up post processing things
	ld_menu_post_timer = vul_timer_create( );
	ld_menu_post_time_modulo_second = 0.f;
	ld_menu_post_last_time = vul_timer_get_micros( ld_menu_post_timer );
	ld_menu_post_phase_rising = SL_TRUE;

	// Creatte the post program
	post_program = sl_renderer_allocate_program( );
	post_program->vert_prog_src = sl_program_default_post_vp_src;
	post_program->frag_prog_src = sl_program_default_post_fp_src;
	sl_program_create( post_program );

	// Create the scene
	s = sl_renderer_add_scene( win->window_id, post_program->program_id );
	
	flip.x = SL_FALSE; flip.y = SL_FALSE;
	color[ 0 ] = color[ 1 ] = color[ 2 ] = color[ 3 ] = 1.f;
	sl_bset_scalar( &uvs, 0.f, 0.f, 1.f, 1.f );
	sl_vset( &pos, 0.0f, 0.0f );
	sl_vset( &scale, 1.f, 1.f );
	sl_bset_scalar( &aabb, -1.f, -1.f, 1.f, 1.f );
	// Background
	t = sl_renderer_allocate_texture( );
	assert( !ld_load_image_to_texture( t, "../res/menu/menu_bg.png" ) );
	sl_scene_add_sprite( s, LD_MENU_BACKGROUND_LAYER,
						 &pos, &scale, 0.f,
						 t->texture_id, 
						 ld_program_instance_id, 
						 ld_renderable_instance_id,
						 &uvs, &flip, color, SL_FALSE );
	
	// Help text, hidden by default
	t = sl_renderer_allocate_texture( );
	assert( !ld_load_image_to_texture( t, "../res/menu/help.png" ) );
	ld_menu_help_id = sl_scene_add_sprite( s, LD_MENU_INFO_LAYER,
										   &pos, &scale, 0.f,
										   t->texture_id, 
										   ld_program_instance_id, 
										   ld_renderable_instance_id,
										   &uvs, &flip, color, SL_TRUE );
	sl_controller_add_mouse_up_callback( s->scene_id, ld_menu_help_id, ld_toggle_help );

	// Credits text, hidden by default
	t = sl_renderer_allocate_texture( );
	assert( !ld_load_image_to_texture( t, "../res/menu/credits.png" ) );
	ld_menu_credits_id = sl_scene_add_sprite( s, LD_MENU_INFO_LAYER,
											  &pos, &scale, 0.f,
											  t->texture_id, 
											  ld_program_instance_id, 
											  ld_renderable_instance_id,
											  &uvs, &flip, color, SL_TRUE );
	sl_controller_add_mouse_up_callback( s->scene_id, ld_menu_credits_id, ld_toggle_credits );
	
	// Background
	t = sl_renderer_allocate_texture( );
	assert( !ld_load_image_to_texture( t, "../res/menu/menu_bg.png" ) );
	sl_scene_add_sprite( s, LD_MENU_BACKGROUND_LAYER,
						 &pos, &scale, 0.f,
						 t->texture_id, 
						 ld_program_instance_id, 
						 ld_renderable_instance_id,
						 &uvs, &flip, color, SL_FALSE );
	// Buttons
	sl_vset( &scale, 0.1f, 0.1f );
	sl_vset( &pos, -0.33f, 0.15f );
	t = sl_renderer_allocate_texture( );
	assert( !ld_load_image_to_texture( t, "../res/menu/menu_play.png" ) );
	ld_menu_play_button = sl_scene_add_sprite( s, LD_MENU_BUTTON_LAYER,
											   &pos, &scale, 0.f,
											   t->texture_id, 
											   ld_program_instance_id, 
											   ld_renderable_instance_id,
											   &uvs, &flip, color, SL_FALSE );
	sl_controller_add_mouse_up_callback( s->scene_id, ld_menu_play_button, ld_play );

	sl_vset( &pos, 0.33f, 0.15f );
	t = sl_renderer_allocate_texture( );
	assert( !ld_load_image_to_texture( t, "../res/menu/menu_quit.png" ) );
	id = sl_scene_add_sprite( s, LD_MENU_BUTTON_LAYER,
							  &pos, &scale, 0.f,
							  t->texture_id, 
							  ld_program_instance_id, 
							  ld_renderable_instance_id,
							  &uvs, &flip, color, SL_FALSE );
	sl_controller_add_mouse_up_callback( s->scene_id, id, ld_quit_mouse );
	
	if( ld_game_custom_level_path ) {
		sl_vset( &pos, 0.f, 0.15f );
		t = sl_renderer_allocate_texture( );
		assert( !ld_load_image_to_texture( t, "../res/menu/menu_custom.png" ) );
		id = sl_scene_add_sprite( s, LD_MENU_BUTTON_LAYER,
								  &pos, &scale, 0.f,
								  t->texture_id, 
								  ld_program_instance_id, 
								  ld_renderable_instance_id,
								  &uvs, &flip, color, SL_FALSE );
		sl_controller_add_mouse_up_callback( s->scene_id, id, ld_play_custom );
	}

	// Help
	sl_vset( &pos, -0.15f, -0.25f );
	t = sl_renderer_allocate_texture( );
	assert( !ld_load_image_to_texture( t, "../res/menu/menu_help.png" ) );
	id = sl_scene_add_sprite( s, LD_MENU_BUTTON_LAYER,
								&pos, &scale, 0.f,
								t->texture_id, 
								ld_program_instance_id, 
								ld_renderable_instance_id,
								&uvs, &flip, color, SL_FALSE );
	sl_controller_add_mouse_up_callback( s->scene_id, id, ld_toggle_help );
	
	// Credits
	sl_vset( &pos, 0.15f, -0.25f );
	t = sl_renderer_allocate_texture( );
	assert( !ld_load_image_to_texture( t, "../res/menu/menu_credits.png" ) );
	id = sl_scene_add_sprite( s, LD_MENU_BUTTON_LAYER,
								&pos, &scale, 0.f,
								t->texture_id, 
								ld_program_instance_id, 
								ld_renderable_instance_id,
								&uvs, &flip, color, SL_FALSE );
	sl_controller_add_mouse_up_callback( s->scene_id, id, ld_toggle_credits );

	// Load music
	ar = sl_renderer_get_aurator_for_scene( s->scene_id );
	clip = sl_aurator_get_clip(ar, sl_aurator_allocate_clip( ar ) );
	sl_aurator_load_ogg( ar, &clip, "../res/audio/menu.ogg" );
	clip->state = SL_AURATOR_RUNNING_LOOPED;

	ld_menu_scene_id = s->scene_id;
	ld_active_scene_id = s->scene_id;

	ld_last_world_scene_id = 0xffffffff;

	ld_game_current_level = 0;
}
