#include "world_loader.h"
#include "ld30.h"
#include "bugger_loader.h"

void ld_world_load_file( sl_scene *scene, const char *path )
{
	FILE *f;
	char line[ 256 ];
	char post_path[ 256 ];
	char *full_file, *c, *cl, *lc;
	size_t file_len, read_len;
	ui32_t w, h, **line_ids, x, y, line_num, bugger_count, bugger_spawn_pause, trampolines;
	sl_vec scale;
	
	// Open the listing
	f = fopen( path, "r" );
	
	// Find file length
	fseek( f, 0, SEEK_END );
	file_len = ftell( f );
	fseek( f, 0, SEEK_SET );
	
	// Read the entire file
	full_file = ( char* )malloc( file_len + 1 );
	read_len = fread( full_file, 1, file_len, f );
	full_file[ file_len ] = 0;
	if( read_len != file_len ) {
		printf( "ld_world_load_file @ File length and read length disagree. %u vs %u.\n", file_len, read_len );
	}
	full_file[ read_len ] = 0;
	fclose( f );
	
	// Scan the file
	c = full_file;
	cl = c;
	line_num = 0;
	line_ids = NULL;
	while( *c ) {
		if( *c == '\n' ) {
			// Scan the line
			strncpy( line, cl, c - cl );
			line[ c - cl ] = 0;
			if( line[ 0 ] != '#' ) { // Skip comments
				if( line_num == 0 ) {
					// First line is size
					sscanf( line, "%u|%u", &w, &h );
					line_ids = ( ui32_t** )malloc( sizeof( ui32_t* ) * h );
					y = 0;
				} else if( line_num == 1 ) {
					sscanf( line, "%u|%u|%u", &bugger_count, &bugger_spawn_pause, &trampolines );
				} else if( line_num == 2 ) {
					snprintf( post_path, 255, "%s", line );
				} else {
					lc = line;
					x = 0;
					line_ids[ y ] = ( ui32_t* )malloc( sizeof( ui32_t* ) * w );
					while( *lc ) {
						if( isdigit( *lc ) ) {
							long val = strtol( lc, &lc, 10 );
							line_ids[ y ][ x++ ] = val;
						} else { // Otherwise, move on to the next character.
							lc++;
						}
					}
					++y;
				}
				++line_num;
			}
			cl = c + 1;
		}
		++c;
	}

	/* Load the scene tiles */
	if( line_ids )
	{
		ld_world_load_array( scene, line_ids, w, h, trampolines );
	} else {
		assert( SL_FALSE ); //Level is malformed
	}

	//ld_world_ui_load( scene, w, h ); //@TODO: Redo this for a relevant menu!

	/* Load the buggers */
	sl_vset( &scale, 1.f / ( f32_t )w, 1.f / ( f32_t )h );
	ld_buggers_load( scene, bugger_count, bugger_spawn_pause, &scale );
	
	/* Load the post-process shader */
	ld_world_load_post_program( scene, post_path );
}

void ld_world_finalize( )
{
	if( ld_world_state_instance )
	{
		free( ld_world_state_instance->tiles );
		free( ld_world_state_instance->trampoline_ids );
		free( ld_world_state_instance );
		ld_world_state_instance = NULL;
	}
}

void ld_world_reload( sl_scene *scene, const char *path )
{
	ui32_t post_id;

	// Unload it
	ld_world_finalize( );

	// For now, just forget the old scene existed, we're moving on to a new one. Who cares 
	// about memory these days anyway...For the small test/tutorial scene, it's 192kb every reset.
	// @TODO: Put this back in and figure out what the hell happens to the menu...
	post_id = scene->post_program_id;
	sl_renderer_finalize_scene( scene->scene_id );
	scene = sl_renderer_add_scene( ld_window_instance_id, post_id );
	ld_buggers_finalize( );

	// Reload it
	ld_world_load_file( scene, path );

	// Set it as active
	ld_active_scene_id = scene->scene_id;

	// Start the buggers
	ld_bugger_start( );
}

void ld_world_load_array( sl_scene *scene, ui32_t **arrays, ui32_t width, ui32_t height, ui32_t trampolines )
{
	ui32_t x, y, swaps;
	f32_t freq;
	sl_vec pos, delta, scale;
	sl_box uvs;
	sl_bvec flip;
	sl_texture *tex;
	sl_animator *a;
	vul_vector_t *anim_states;
	GLfloat color[ 4 ];
	unsigned int id;
	
	// Initiate world array
	assert( ld_world_state_instance == NULL );
	ld_world_state_instance = ( ld_world_state* )malloc( sizeof( ld_world_state ) );
	ld_world_state_instance->width = width;
	ld_world_state_instance->height = height;
	ld_world_state_instance->tiles = ( ld_world_tile* )malloc( sizeof( ld_world_tile ) * width * height );
	ld_world_state_instance->trampolines_left = trampolines;
	ld_world_state_instance->trampoline_ids = ( ui32_t* )malloc( sizeof( ui32_t ) * ld_world_state_instance->trampolines_left );
	
	color[ 0 ] = 1.f; color[ 1 ] = 1.f; color[ 2 ] = 1.f; color[ 3 ] = 1.f;
	sl_bset_scalar( &uvs, 0.f, 0.f, 1.f, 1.f );
	sl_vset( &delta, 2.f / ( f32_t )width, 2.f / ( f32_t )height );
	sl_vmuls( &scale, &delta, 0.5f ); // Scale is [0,1], not [-1,1], so half of delta
	sl_vset( &pos, -1.f + scale.x, 1.f - scale.y );
	flip.x = SL_FALSE; flip.y = SL_FALSE;
	for( y = 0; y < height; ++y ) {
		pos.x = -1.f + scale.x;
		for( x = 0; x < width; ++x ) {
			ld_tile_dictionary_get( &tex, &anim_states, &uvs, &freq, &swaps, arrays[ y ][ x ] );
			assert( tex != NULL ); // Invalid tile!
			// Add the sprite
			id = sl_scene_add_sprite( scene, LD_WORLD_TILE_LAYER,
									  &pos, &scale, 0.f,
									  tex->texture_id, 
									  ld_program_instance_id, 
									  ld_renderable_instance_id,
									  &uvs, &flip, color, SL_FALSE );
			// If there's an animation, add it
			if( anim_states != NULL ){
				a = sl_renderer_get_animator_for_scene( scene->scene_id );
				sl_animator_add_sprite( a, id, anim_states, ( ui64_t )( freq * 1000.0 ), SL_ANIMATION_RUNNING_LOOPED );
			}
			sl_controller_add_mouse_over_callback( scene->scene_id, id, ld_world_tile_over );
			sl_controller_add_mouse_out_callback( scene->scene_id, id, ld_world_tile_out );
			sl_controller_add_mouse_up_callback( scene->scene_id, id, ld_world_tile_swap );
			// Finally store thee world state
			ld_world_state_instance->tiles[ y * width + x ].tile_type = arrays[ y ][ x ];
			ld_world_state_instance->tiles[ y * width + x ].swaps_left = swaps;
			ld_world_state_instance->tiles[ y * width + x ].quad_id = id;
			// And if this is entry, warp or exit, store it's location.
			if( arrays[ y ][ x ] == 8 ) {
				ld_world_indices_to_normalized_coords( &ld_world_state_instance->entry, x, y );
			} else if( arrays[ y ][ x ] == 10 ) {
				ld_world_indices_to_normalized_coords( &ld_world_state_instance->warp_out, x, y );
			}
			// And move along position
			pos.x += delta.x;
		}
		pos.y -= delta.y;
	}

	/* Load the trampoline icons in the bottom left. These are half-sized and not animated */
	ld_tile_dictionary_get( &tex, &anim_states, &uvs, &freq, &swaps, LD_WORLD_TILE_TYPE_TRAMPOLINE );
	sl_vmuls( &scale, &scale, 0.5f );
	uvs.min_p.x = uvs.max_p.x * 0.5f;
	uvs.min_p.y = uvs.max_p.y * 0.5f;
	sl_vset( &pos, -1.f + scale.x * 1.1f, -1.f + scale.x * 1.1f );
	for( x = 0; x < ld_world_state_instance->trampolines_left; ++x ) {
		ld_world_state_instance->trampoline_ids[ x ]
			   = sl_scene_add_sprite( scene, LD_WORLD_OVERLAY_LAYER,
									  &pos, &scale, 0.f,
									  tex->texture_id, 
									  ld_program_instance_id, 
									  ld_renderable_instance_id,
									  &uvs, &flip, color, SL_FALSE );
		pos.x += scale.x * 2.2f;
	}

}

void ld_world_load_post_program( sl_scene *scene, const char *post_path )
{
	FILE *f;
	char *fstr;
	size_t flen, rlen;
	sl_program *p;
	
	// Open the listing
	f = fopen( post_path, "r" );
	
	// Find file length
	fseek( f, 0, SEEK_END );
	flen = ftell( f );
	fseek( f, 0, SEEK_SET );
	
	// Read the entire file
	fstr = ( char* )malloc( flen + 1 );
	rlen = fread( fstr, 1, flen, f );
	fstr[ flen ] = 0;
	if( rlen != flen ) {
		printf( "ld_world_load_post_program @ File length and read length disagree. %u vs %u.\n", flen, rlen );
	}
	fstr[ rlen ] = 0;
	fclose( f );

	// Create the program	
	p = sl_renderer_allocate_program( );
	p->vert_prog_src = sl_program_default_post_vp_src;
	p->frag_prog_src = fstr;
	sl_program_create( p );

	// Add it to the scene
	sl_scene_set_post( scene, p, NULL );

	// @TODO: Deal with the call back somehow!
}

void ld_world_indices_to_normalized_coords( sl_vec *normalized_pos, ui32_t ix, ui32_t iy )
{
	sl_vec delta;
	sl_vec offset;
	
	// Make sure we are within the world bounds.
	assert( ix >= 0 && ix < ld_world_state_instance->width
		 && iy >= 0 && iy < ld_world_state_instance->height );

	sl_vset( &delta, 2.f / ( f32_t )ld_world_state_instance->width,
					 2.f / ( f32_t )ld_world_state_instance->height );
	sl_vmuls( &offset, &delta, 0.5f );
	
	sl_vset( normalized_pos, -1.f +   delta.x * ( f32_t )ix + offset.x,
							  1.f - ( delta.y * ( f32_t )iy + offset.y ) );
}

void ld_world_normalized_coords_to_indices( ui32_t *ix, ui32_t *iy, sl_vec *normalized_pos )
{
	sl_vec pos_zero_to_one;
	sl_vec offset;
	sl_vec tmp;

	// Assert it is in fact noramlized; otherwise we will get very, very wrong indices.
	assert( -1.0000001f < normalized_pos->x && normalized_pos->x < 1.0000001f
		 && -1.0000001f < normalized_pos->y && normalized_pos->y < 1.0000001f );

	sl_vset( &tmp, 0.5f, 0.5f );
	sl_vmul( &pos_zero_to_one, normalized_pos, &tmp );
	sl_vadd( &pos_zero_to_one, &pos_zero_to_one, &tmp );
	sl_vset( &offset, 1.f / ( f32_t )ld_world_state_instance->width,
					  1.f / ( f32_t )ld_world_state_instance->height );
	
	*ix = ( ui32_t )( ( f32_t )ld_world_state_instance->width * pos_zero_to_one.x );
	*iy = ( ui32_t )( ( f32_t )ld_world_state_instance->height * ( 1.f - pos_zero_to_one.y ) ); // OpenGL being inverted again...
}

ui32_t ld_world_swaps_remaining( sl_vec *normalized_pos )
{
	ui32_t ix, iy;
	
	ld_world_normalized_coords_to_indices( &ix, &iy, normalized_pos );

	return ld_world_state_instance->tiles[ iy * ld_world_state_instance->width + ix ].swaps_left;
}

ui32_t ld_world_get_quad_id_for_pos( sl_vec *normalized_pos )
{
	ui32_t ix, iy;
	
	ld_world_normalized_coords_to_indices( &ix, &iy, normalized_pos );

	return ld_world_state_instance->tiles[ iy * ld_world_state_instance->width + ix ].quad_id;
}

void ld_world_tile_over( unsigned int scene_id, unsigned int quad_id )
{
	sl_quad *q;
	sl_vec pos, p2;
	sl_scene *s;
	ui32_t swaps, qid, ix, iy, mid;

	if( ld_active_scene_id != scene_id ) {
		return;
	}
	s = sl_renderer_get_scene_by_id( ld_active_scene_id );
	q = sl_scene_get_volitile_quad( s, quad_id, LD_WORLD_TILE_LAYER );
	sl_vset( &pos, q->world_matrix.data[ 12 ], q->world_matrix.data[ 13 ] );
	swaps = ld_world_swaps_remaining( &pos );

	if( swaps == 0 ) {
		q->color[ 0 ] = 1.3f;
		q->color[ 1 ] = 1.f;
	} else if( swaps == 1 ) {
		q->color[ 0 ] = 1.2f;
		q->color[ 1 ] = 1.2f;
	} else {
		q->color[ 0 ] = 1.f;
		q->color[ 1 ] = 1.3f;
		q->color[ 2 ] = 1.2f;
	}
	
	// Also update the corresponing swapping tile
	ld_world_normalized_coords_to_indices( &ix, &iy, &pos );
	mid = ( ( ld_world_state_instance->height + 1 ) / 2 ) - 1;
	if( iy > mid ) {
		iy = iy - ( mid + 1 );
	} else if( iy < mid ) {
		iy = iy + ( mid + 1 );
	} else {
		// Can't swap with self,
		return;
	}
	ld_world_indices_to_normalized_coords( &p2, ix, iy );

	qid = ld_world_get_quad_id_for_pos( &p2 );
	q = sl_scene_get_volitile_quad( s, qid, LD_WORLD_TILE_LAYER );
	sl_vset( &pos, q->world_matrix.data[ 12 ], q->world_matrix.data[ 13 ] );
	swaps = ld_world_swaps_remaining( &pos );
	
	if( swaps == 0 ) {
		q->color[ 0 ] = 1.3f;
		q->color[ 1 ] = 1.f;
	} else if( swaps == 1 ) {
		q->color[ 0 ] = 1.2f;
		q->color[ 1 ] = 1.2f;
	} else {
		q->color[ 0 ] = 1.f;
		q->color[ 1 ] = 1.3f;
		q->color[ 2 ] = 1.2f;
	}
}

void ld_world_tile_out( unsigned int scene_id, unsigned int quad_id )
{
	sl_quad *q;
	ui32_t ix, iy, mid, qid;
	sl_vec pos, p2;
	sl_scene *s;
	
	if( ld_active_scene_id != scene_id ) {
		return;
	}
	
	s = sl_renderer_get_scene_by_id( ld_active_scene_id );
	q = sl_scene_get_volitile_quad( s, quad_id, LD_WORLD_TILE_LAYER );
	sl_vset( &pos, q->world_matrix.data[ 12 ], q->world_matrix.data[ 13 ] );

	q->color[ 0 ] = 1.0f;
	q->color[ 1 ] = 1.0f;
	q->color[ 2 ] = 1.0f;

	// Also update the corresponing swapping tile
	ld_world_normalized_coords_to_indices( &ix, &iy, &pos );
	mid = ( ( ld_world_state_instance->height + 1 ) / 2 ) - 1;
	if( iy > mid ) {
		iy = iy - ( mid + 1 );
	} else if( iy < mid ) {
		iy = iy + ( mid + 1 );
	} else {
		// Can't swap with self,
		return;
	}
	ld_world_indices_to_normalized_coords( &p2, ix, iy );

	qid = ld_world_get_quad_id_for_pos( &p2 );
	q = sl_scene_get_volitile_quad( s, qid, LD_WORLD_TILE_LAYER );
	
	q->color[ 0 ] = 1.0f;
	q->color[ 1 ] = 1.0f;
	q->color[ 2 ] = 1.0f;
}

void ld_world_swap( sl_scene *scene, sl_vec *normalized_pos )
{
	ui32_t ix, iy, iyb, uitmp, uitmp2, mid, bix, biy;
	const sl_quad *qbugger;
	sl_vec pbug;

	ld_world_normalized_coords_to_indices( &ix, &iy, normalized_pos );
	
	// Find the other quad
	mid = ( ( ld_world_state_instance->height + 1 ) / 2 ) - 1;
	if( iy > mid ) {
		iyb = iy - ( mid + 1 );
	} else if( iy < mid ) {
		iyb = iy + ( mid + 1 );
	} else {
		// Can't swap with self,
		return;
	}
	
	if(    ld_world_state_instance->tiles[ iy  * ld_world_state_instance->width + ix ].swaps_left <= 0
		|| ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ix ].swaps_left <= 0
		|| ld_world_state_instance->tiles[ iy * ld_world_state_instance->width + ix ].swaps_left == TILE_PASSIVLEY_SWAPPABLE )
	{
		printf( "No more swaps!\n" ); // @TODO: play sound/alert player to illeaglity!
		return;
	}

	// Check if we are swapping ground for a wall; that doesn't do anything!
	if(		 ( ld_world_state_instance->tiles[ iy  * ld_world_state_instance->width + ix ].tile_type == LD_WORLD_TILE_TYPE_WALL
			&& ld_world_state_instance->tiles[ iyb  * ld_world_state_instance->width + ix ].tile_type == LD_WORLD_TILE_TYPE_GROUND )
		  || ( ld_world_state_instance->tiles[ iy  * ld_world_state_instance->width + ix ].tile_type == LD_WORLD_TILE_TYPE_GROUND
			&& ld_world_state_instance->tiles[ iyb  * ld_world_state_instance->width + ix ].tile_type == LD_WORLD_TILE_TYPE_WALL ) )
	{
		printf( "Swapped tiles, but no really (doesn't do anything)\n" );
		return;
	}

	// Check there are no buggers there
	for( uitmp = 0; uitmp < ld_bugger_handler_instance->bugger_count; ++uitmp )
	{
		qbugger = sl_scene_get_const_quad( scene, ld_bugger_handler_instance->buggers[ uitmp ].quad_id, LD_WORLD_BUGGER_LAYER );
		sl_vset( &pbug, qbugger->world_matrix.data[ 12 ], qbugger->world_matrix.data[ 13 ] );
		ld_world_normalized_coords_to_indices( &bix, &biy, &pbug );

		if( bix == ix && ( biy == iy || biy == iyb ) ) {
			printf( "No swapping while tile is busy!" );
			return; 
		}
	}

	// Swap the tiles
	ld_world_swap_tiles( scene, ix, iy, ix, iyb );

	// If you just swapped earth/static into a space where the next above it is a wall,
	// Make that wall into ground.
	if( iyb > 0 ) {
		uitmp = ld_world_state_instance->tiles[ ( iyb - 1 )  * ld_world_state_instance->width + ix ].tile_type;
		uitmp2 = ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ix ].tile_type;
		if( ( uitmp2 == LD_WORLD_TILE_TYPE_EARTH || uitmp2 == LD_WORLD_TILE_TYPE_STATIC )
			&& uitmp == LD_WORLD_TILE_TYPE_WALL ) {
			ld_world_change_tile( scene, ix, iyb - 1, LD_WORLD_TILE_TYPE_GROUND );
		}
	}
	// If you just swapped ground or wall just below ground, remove ground abvoe
	if( iyb > 0 ) {
		uitmp = ld_world_state_instance->tiles[ ( iyb - 1 )  * ld_world_state_instance->width + ix ].tile_type;
		uitmp2 = ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ix ].tile_type;
		if( ( uitmp2 == LD_WORLD_TILE_TYPE_WALL || uitmp2 == LD_WORLD_TILE_TYPE_GROUND )
			&& uitmp == LD_WORLD_TILE_TYPE_GROUND ) {
			ld_world_change_tile( scene, ix, iyb - 1, LD_WORLD_TILE_TYPE_WALL );
		}
	}
	// Repeat for the one we swapped from
	if( iy > 0 ) {
		uitmp = ld_world_state_instance->tiles[ ( iy - 1 )  * ld_world_state_instance->width + ix ].tile_type;
		uitmp2 = ld_world_state_instance->tiles[ iy * ld_world_state_instance->width + ix ].tile_type;
		if( ( uitmp2 == LD_WORLD_TILE_TYPE_EARTH || uitmp2 == LD_WORLD_TILE_TYPE_STATIC )
			&& uitmp == LD_WORLD_TILE_TYPE_WALL ) {
			ld_world_change_tile( scene, ix, iy - 1, LD_WORLD_TILE_TYPE_GROUND );
		}
	}
	if( iy > 0 ) {
		uitmp = ld_world_state_instance->tiles[ ( iy - 1 )  * ld_world_state_instance->width + ix ].tile_type;
		uitmp2 = ld_world_state_instance->tiles[ iy * ld_world_state_instance->width + ix ].tile_type;
		if( ( uitmp2 == LD_WORLD_TILE_TYPE_WALL || uitmp2 == LD_WORLD_TILE_TYPE_GROUND )
			&& uitmp == LD_WORLD_TILE_TYPE_GROUND ) {
			ld_world_change_tile( scene, ix, iy - 1, LD_WORLD_TILE_TYPE_WALL );
		}
	}
	// If we just swapped ground above ground, make top one a wall
	if( iy < ld_world_state_instance->height ) {
		if( ld_world_state_instance->tiles[ iy * ld_world_state_instance->width + ix ].tile_type == LD_WORLD_TILE_TYPE_GROUND
			&& ld_world_state_instance->tiles[ ( iy + 1 ) * ld_world_state_instance->width + ix ].tile_type == LD_WORLD_TILE_TYPE_GROUND )
		{
			ld_world_change_tile( scene, ix, iy, LD_WORLD_TILE_TYPE_WALL );
		}
	}
	if( iyb < ld_world_state_instance->height ) {
		if( ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ix ].tile_type == LD_WORLD_TILE_TYPE_GROUND
			&& ld_world_state_instance->tiles[ ( iyb + 1 ) * ld_world_state_instance->width + ix ].tile_type == LD_WORLD_TILE_TYPE_GROUND )
		{
			ld_world_change_tile( scene, ix, iyb, LD_WORLD_TILE_TYPE_WALL );
		}
	}

}

void ld_world_tile_swap( unsigned int scene_id, unsigned quad_id, int button )
{
	sl_quad *q;
	sl_vec pos;
	sl_scene *s;

	if( ld_active_scene_id != scene_id ) {
		return;
	}
	s = sl_renderer_get_scene_by_id( ld_active_scene_id );
	q = sl_scene_get_volitile_quad( s, quad_id, LD_WORLD_TILE_LAYER );

	sl_vset( &pos, q->world_matrix.data[ 12 ], q->world_matrix.data[ 13 ] );

	if( button == GLFW_MOUSE_BUTTON_1 ) {
		ld_world_swap( s, &pos );
	} else {
		ld_world_insert_trampoline( s, &pos );
	}
}

void ld_world_change_tile( sl_scene *scene, ui32_t xi, ui32_t yi, ui32_t tile_id )
{
	sl_texture *tex;
	vul_vector_t *as;
	sl_box uvs;
	f32_t freq;
	ui32_t swaps;
	sl_quad *q;
	ld_world_tile *tile;
	sl_animator *a;

	ld_tile_dictionary_get( &tex, &as, &uvs, &freq, &swaps, tile_id );

	tile = &ld_world_state_instance->tiles[ yi * ld_world_state_instance->width + xi ];
	tile->tile_type = tile_id;

	q = sl_scene_get_volitile_quad( scene, tile->quad_id, LD_WORLD_TILE_LAYER );
	q->texture_id = tex->texture_id;
	
	switch( tile_id )
	{
	case LD_WORLD_TILE_TYPE_WALL:
	case LD_WORLD_TILE_TYPE_EARTH:
	case LD_WORLD_TILE_TYPE_GROUND:
	case LD_WORLD_TILE_TYPE_POROUS:
	case LD_WORLD_TILE_TYPE_STATIC:
	case LD_WORLD_TILE_TYPE_SPIKES:
	case LD_WORLD_TILE_TYPE_EARTH_SINGLE_SWAP:
	case LD_WORLD_TILE_TYPE_ENTRY:
	case LD_WORLD_TILE_TYPE_WARP_IN:
	case LD_WORLD_TILE_TYPE_WARP_OUT:
	case LD_WORLD_TILE_TYPE_EXIT:
		q->uvs.max_p.x = 1.f;		
		q->uvs.max_p.y = 1.f;
		break; // Just make sure we have full uvs.
	case LD_WORLD_TILE_TYPE_TRAMPOLINE:
		// Now we need to update animation as well.
		q->uvs.max_p.x = 0.5f;
		q->uvs.max_p.y = 0.5f;
		a = sl_renderer_get_animator_for_scene( scene->scene_id );
		sl_animator_add_sprite( a, q->quad_id, as, ( ui64_t )( freq * 1000.f ), SL_ANIMATION_RUNNING_LOOPED );
		break;
	}
}

void ld_world_swap_tiles( sl_scene *scene, ui32_t ix, ui32_t iy, ui32_t ixb, ui32_t iyb )
{
	ui32_t qai, qbi, uitmp;
	sl_quad *qa, *qb;
	f32_t tmp_x, tmp_y;

	// Find quads
	qai = ld_world_state_instance->tiles[ iy  * ld_world_state_instance->width + ix ].quad_id;
	qbi = ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ixb ].quad_id;

	// Swap them in world representation
	ld_world_state_instance->tiles[ iy  * ld_world_state_instance->width + ix ].quad_id = qbi;
	ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ixb ].quad_id = qai;

	// Swap their types
	uitmp = ld_world_state_instance->tiles[ iy  * ld_world_state_instance->width + ix ].tile_type;
	ld_world_state_instance->tiles[ iy  * ld_world_state_instance->width + ix ].tile_type
		= ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ixb ].tile_type;
	ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ixb ].tile_type  = uitmp;
	
	// Decrease and swap swaps left
	uitmp = ld_world_state_instance->tiles[ iy  * ld_world_state_instance->width + ix ].swaps_left - 1;
	if( ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ixb ].swaps_left != TILE_INFINITELY_SWAPPABLE
		&& ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ixb ].swaps_left != TILE_PASSIVLEY_SWAPPABLE )
	{
		ld_world_state_instance->tiles[ iy  * ld_world_state_instance->width + ix ].swaps_left
			= ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ixb ].swaps_left - 1;
	}
	if( uitmp + 1 != TILE_INFINITELY_SWAPPABLE && uitmp + 1 != TILE_PASSIVLEY_SWAPPABLE ) {
		ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ixb ].swaps_left = uitmp;
	}

	// Change their rendering position
	qa = sl_scene_get_volitile_quad( scene, qai, LD_WORLD_TILE_LAYER );
	qb = sl_scene_get_volitile_quad( scene, qbi, LD_WORLD_TILE_LAYER );

	tmp_x = qa->world_matrix.data[ 12 ];
	tmp_y = qa->world_matrix.data[ 13 ];
	qa->world_matrix.data[ 12 ] = qb->world_matrix.data[ 12 ];
	qa->world_matrix.data[ 13 ] = qb->world_matrix.data[ 13 ];
	qb->world_matrix.data[ 12 ] = tmp_x;
	qb->world_matrix.data[ 13 ] = tmp_y;
	
	// Also change their color, since the mouse over state is now messed up
	qa->color[ 0 ] = 1.f; qa->color[ 1 ] = 1.f; qa->color[ 2 ] = 1.f;
	if( ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ix ].swaps_left == 0 ) {
		qb->color[ 0 ] = 1.3f; qb->color[ 1 ] = 1.f; qb->color[ 2 ] = 1.f;
	} else if (ld_world_state_instance->tiles[ iyb * ld_world_state_instance->width + ix ].swaps_left == 1 ) {
		qb->color[ 0 ] = 1.2f; qb->color[ 1 ] = 1.2f; qb->color[ 2 ] = 1.f;
	} else {
		qb->color[ 0 ] = 1.f; qb->color[ 1 ] = 1.3f; qb->color[ 2 ] = 1.2f;
	}

	printf( "Swapped %d with %d\n", qai, qbi );
}

void ld_world_insert_trampoline( sl_scene *scene, sl_vec *normalized_pos )
{
	ui32_t ix, iy, type;
	sl_quad *q;

	ld_world_normalized_coords_to_indices( &ix, &iy, normalized_pos );
	type = ld_world_state_instance->tiles[ iy * ld_world_state_instance->width + ix ].tile_type;
	if( type == LD_WORLD_TILE_TYPE_GROUND && ld_world_state_instance->trampolines_left ) {
		ld_world_change_tile( scene, ix, iy, LD_WORLD_TILE_TYPE_TRAMPOLINE );
		q = sl_scene_get_volitile_quad( scene,
				ld_world_state_instance->trampoline_ids[ --ld_world_state_instance->trampolines_left ],
				LD_WORLD_OVERLAY_LAYER );
		q->hidden = SL_TRUE;
	}
}
